default_shortcuts = {
    'napari:toggle_console_visibility': ['Control-Shift-C'],
    'napari:reset_scroll_progress': ['Control'],
    'napari:toggle_ndisplay': ['Control-Y'],
    'napari:toggle_theme': ['Control-Shift-T'],
    'napari:reset_view': ['Control-R'],
    'napari:increment_dims_left': ['Left'],
    'napari:increment_dims_right': ['Right'],
    'napari:focus_axes_up': ['Alt-Up'],
    'napari:focus_axes_down': ['Alt-Down'],
    'napari:roll_axes': ['Control-E'],
    'napari:transpose_axes': ['Control-T'],
    'napari:toggle_grid': ['Control-G'],
    'napari:toggle_selected_visibility': ['V'],
    'napari:activate_label_erase_mode': ['1'],
    'napari:activate_fill_mode': ['3'],
    'napari:activate_paint_mode': ['2'],
    'napari:activate_label_pan_zoom_mode': ['5'],
    'napari:activate_label_picker_mode': ['4'],
    'napari:new_label': ['M'],
    'napari:decrease_label_id': ['-'],
    'napari:increase_label_id': ['='],
    'napari:activate_points_add_mode': ['2'],
    'napari:activate_points_select_mode': ['3'],
    'napari:activate_points_pan_zoom_mode': ['4'],
    'napari:select_all': ['A'],
    'napari:delete_selected_points': ['Backspace', 'Delete', '1'],
    'napari:activate_add_rectangle_mode': ['R'],
    'napari:activate_add_ellipse_mode': ['E'],
    'napari:activate_add_line_mode': ['L'],
    'napari:activate_add_path_mode': ['T'],
    'napari:activate_add_polygon_mode': ['P'],
    'napari:activate_direct_mode': ['4'],
    'napari:activate_select_mode': ['5'],
    'napari:activate_shape_pan_zoom_mode': ['6'],
    'napari:activate_vertex_insert_mode': ['2'],
    'napari:activate_vertex_remove_mode': ['1'],
    'napari:copy_selected_shapes': ['Control-C'],
    'napari:paste_shape': ['Control-V'],
    'napari:select_all_shapes': ['A'],
    'napari:delete_selected_shapes': ['Backspace', 'Delete', '3'],
    'napari:finish_drawing_shape': ['Escape'],
    'napari:reset_active_layer_affine': ['Control-Shift-R'],
    'napari:transform_active_layer': ['Control-Shift-A'],
}
